# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin literal_cmd user}
{xrst_spell
   dir
   literalinclude
}

Literal Command
###############

Syntax
******

-  ``\{xrst_literal}``

-  | ``\{xrst_literal``
   |     *display_file*
   | ``}``

-  | ``\{xrst_literal``
   |     *start_after*
   |     *end_before*
   | ``}``

-  | ``\{xrst_literal``
   |     *display_file*
   |     *start_after*
   |     *end_before*
   | ``}``

Purpose
*******
A code block, from any where in any file,
can be included by the command above.

literalinclude
**************
This command is similar to the following sphinx directive
(see :ref:`dir_cmd-name`) :

| |tab| .. literalinclude:: \{xrst_dir *display_file*}
| |tab| |tab| :start-after: *start_after*
| |tab| |tab| :end-before: *end_before*

The xrst literal command has the following difference:

#. If the *display_file* is not specified, the current input file is used.
#. The copy of *start_after* and *end_before* in the command is not considered
   a match for the corresponding text. This makes it possible to put the
   command above the text when *display_file* is the current input file.
#. It is an error for there to be more than one copy of *start_after*
   or *end_before* in the *display_file* (not counting the copy in the
   command when the display file is the current input file).
   This makes sure that the intended section of *display_file* is displayed.

White Space
***********
Leading and trailing white space is not included in
*start_after*, *end_before* or *display_file*.
The new line character separates these tokens.
The line containing the ``}`` must have nothing but white space after it.

display_file
************
If *display_file* is not in the syntax,
the code block is in the current input file.
Otherwise, the code block is in *display_file*.
The file name *display_file* is relative to the
:ref:`config_file@directory@project_directory` .

1. This may seem verbose, but it makes it easier to write scripts
   that move files and automatically change references to them.
2. Note that if you use the sphinx ``literalinclude`` directive,
   the corresponding file name will be relative to the
   :ref:`config_file@directory@rst_directory` , which is a path relative
   to the project_directory.

No start or end
***************
In the case where there is no *start_after* or *end_before*,
the entire display file is displayed.
In the case of the ``\{xrst_literal}`` syntax,
the entire current input file is displayed.

start_after
***********
The code block starts with the line following the occurrence
of the text *start_after* in *display_file*.
If this is the same as the file containing the command,
the text *start_after* will not match any text in the command.
There must be one and only one occurrence of *start_after* in *display_file*,
not counting the command itself when the files are the same.

end_before
**********
The code block ends with the line before the occurrence
of the text *end_before* in *display_file*.
If this is the same as the file containing the command,
the text *end_before* will not match any text in the command.
There must be one and only one occurrence of *end_before* in *display_file*,
not counting the command itself when the files are the same.

Spell Checking
**************
Spell checking is **not** done for these code blocks.


Example
*******
see :ref:`literal_example-name` .

{xrst_end literal_cmd}
"""
# ----------------------------------------------------------------------------
import os
import re
import xrst
#
# ----------------------------------------------------------------------------
#
# extension_map
# map cases that pygments has trouble with
extension_map = {
   'xrst' : 'rst'    ,
   'hpp'  : 'cpp'    ,
   'm'    : 'matlab' ,
   'txt'  : ''       ,
}
def file_extension(display_file) :
   index = display_file.rfind('.')
   extension = ''
   if 0 <= index and index + 1 < len(display_file) :
      extension = display_file[index + 1 :]
      if extension in extension_map :
         extension = extension_map[extension]
   return extension
# ----------------------------------------------------------------------------
# {xrst_begin literal_cmd_dev dev}
# {xrst_spell
#     dir
# }
# {xrst_comment_ch #}
#
# Process the literal commands in a page
# ######################################
#
# Arguments
# *********
#
# data_in
# =======
# is the data for a page before the
# :ref:`literal commands <literal_cmd-name>` have been removed.
#
# file_name
# =========
# is the name of the file that this data comes from. This is used
# for error reporting and for the display file (when the display file
# is not included in the command).
#
# page_name
# =========
# is the name of the page that this data is in. This is only used
# for error reporting.
#
# rst2project_dir
# ===============
# is a relative path from the :ref:`config_file@directory@rst_directory`
# to the :ref:`config_file@directory@project_directory` .
#
# Returns
# *******
#
# data_out
# ========
# Each xrst literal command is converted to its corresponding sphinx commands.
#
# {xrst_code py}
def literal_command(data_in, file_name, page_name, rst2project_dir) :
   assert type(data_in) == str
   assert type(file_name) == str
   assert type(page_name) == str
   assert type(rst2project_dir) == str
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end literal_cmd_dev}
   #
   assert xrst.pattern['literal_0'].groups == 1
   assert xrst.pattern['literal_1'].groups == 4
   assert xrst.pattern['literal_2'].groups == 6
   assert xrst.pattern['literal_3'].groups == 8
   #
   # data_out
   data_out = data_in
   #
   # key
   for key in [ 'literal_0', 'literal_1' ] :
      #
      # m_file
      m_file  = xrst.pattern[key].search(data_out)
      while m_file != None :
         #
         # display_file
         if key == 'literal_0' :
            display_file = file_name
         else :
            display_file = m_file.group(2).strip()
            if not os.path.isfile(display_file) :
               msg  = 'literal command: can not find the display_file.\n'
               msg += f'display_file = {display_file}'
               xrst.system_exit(msg,
                  file_name    = file_name,
                  page_name = page_name,
                  m_obj        = m_file,
                  data         = data_out
               )
            if os.path.samefile(display_file, file_name) :
               display_file = file_name
         #
         # cmd
         display_path = os.path.join(rst2project_dir, display_file)
         cmd          = f'.. literalinclude:: {display_path}\n'
         extension = file_extension( display_file )
         if extension != '' :
            cmd += 3 * ' ' + f':language: {extension}\n'
         cmd = '\n' + cmd + '\n'
         if data_out[m_file.start() ] != '\n' :
            cmd = '\n' + cmd
         #
         # data_out
         data_tmp  = data_out[: m_file.start() + 1 ]
         data_tmp += cmd
         data_tmp += data_out[ m_file.end() : ]
         data_out  = data_tmp
         #
         # m_file
         m_file  = xrst.pattern[key].search(data_out)
         if m_file and key == 'literal_0' :
            msg  = 'More than one {xrst_literal} command in this page.\n'
            msg += 'This command includes the entire current input file.'
            xrst.system_exit(msg,
               file_name    = file_name,
               page_name = page_name,
               m_obj        = m_file,
               data         = data_out
            )
   #
   # key
   for key in [ 'literal_2', 'literal_3' ] :
      #
      # m_file
      m_file  = xrst.pattern[key].search(data_out)
      while m_file != None :
         #
         # display_file, start_after, end_before, display_file, cmd_line
         if key == 'literal_2' :
            display_file  = file_name
            start_after   = m_file.group(2).strip()
            end_before    = m_file.group(4) .strip()
            cmd_end_line = int( m_file.group(6) )
         else :
            display_file  = m_file.group(2).strip()
            start_after   = m_file.group(4).strip()
            end_before    = m_file.group(6) .strip()
            cmd_end_line = int( m_file.group(8) )
            if not os.path.isfile(display_file) :
               msg  = 'literal command: can not find the display_file.\n'
               msg += f'display_file = {display_file}'
               xrst.system_exit(msg,
                  file_name    = file_name,
                  page_name = page_name,
                  m_obj        = m_file,
                  data         = data_out
               )
            same_file   = os.path.samefile(display_file, file_name)
            if same_file :
               display_file = file_name
         cmd_start_line = int( m_file.group(1) )
         cmd_line       = (cmd_start_line, cmd_end_line)
         #
         # start_line, end_line
         start_line, end_line = xrst.start_end_file(
            file_cmd     = file_name,
            page_name = page_name,
            display_file = display_file,
            cmd_line     = cmd_line,
            start_after  = start_after,
            end_before   = end_before
         )
         #
         # locations in display_file
         start_line  = start_line + 1
         end_line    = end_line  - 1
         #
         # cmd
         display_path = os.path.join(rst2project_dir, display_file)
         cmd          = f'.. literalinclude:: {display_path}\n'
         cmd         += 3 * ' ' + f':lines: {start_line}-{end_line}\n'
         #
         # cmd
         # Add language to literalinclude, sphinx seems to be brain
         # dead and does not do this automatically.
         extension = file_extension( display_file )
         if extension != '' :
            cmd += 3 * ' ' + f':language: {extension}\n'
         cmd = '\n' + cmd + '\n\n'
         if m_file.start() > 0 :
            if data_out[m_file.start() - 1] != '\n' :
               cmd = '\n' + cmd
         #
         # data_out
         data_tmp  = data_out[: m_file.start() + 1 ]
         data_tmp += cmd
         data_tmp += data_out[ m_file.end() : ]
         data_out  = data_tmp
         #
         # m_file
         m_file  = xrst.pattern[key].search(data_out)
         #
   #
   xrst.check_syntax_error(
      command_name    = 'literal',
      data            = data_out,
      file_name       = file_name,
      page_name    = page_name,
   )
   # BEGIN_return
   assert type(data_out) == str
   #
   return data_out
   # END_return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin toc_cmd user}
{xrst_spell
   frist
   toc
   toctree
}

Table of Children Commands
##########################

Syntax
******

toc_hidden
==========
| ``\{xrst_toc_hidden``
|   *file_1*
|   ...
|   *file_n*
| ``}``

toc_list
========
| ``\{xrst_toc_list``
|   *file_1*
|   ...
|   *file_n*
| ``}``

toc_table
=========
| ``\{xrst_toc_table``
|   *file_1*
|   ...
|   *file_n*
| ``}``

Table of Contents
*****************
These commands specify the pages that are children
of the current page; i.e., pages that are at the
next level in the table of contents.

File Names
**********
A new line character must precede and follow each
of the file names *file_1* ... *file_n*.
Leading and trailing white space is not included in the names
The file names are  relative to the
:ref:`config_file@directory@project_directory` .
This may seem verbose, but it makes it easier to write scripts
that move files and automatically change references to them.

Children
********
Each of the files may contain multiple :ref:`pages<begin_cmd@Page>`.
The first of these pages may use a
:ref:`parent begin<begin_cmd@Parent Page>` command.

#. The first page in a file is always a child of the
   page where the toc command appears..

#. If the first page in a file is a begin parent page,
   the other pages in the file are children of the frist page.
   Hence the other pages are grand children of the page
   where the begin toc command appears.

#. If there is no begin parent command in a file,
   all the pages in the file are children of the
   page where the toc command appears.

#. If the first page in a file is a begin parent page,
   and there is also a toc command in this page,
   links to the toc command children come first and then links to
   the children that are other pages in the same file.

Child Links
***********
#. The toc_list syntax generates links to the children that
   display the title for each page.
   The toc_table syntax generates links to the children that
   display both the page name and page tile.

#. If a page has a toc_list or toc_table command,
   links to all the children of the page are placed where the
   toc command is located.
   You can place a heading directly before these commands
   to make the links easier to find.

#. If a page uses the hidden syntax,
   no automatic links to the children of the current page are generated.

#. If a page does not have a toc command,
   and it has a begin parent command,
   links to the children of the page are placed at the end of the page.

toctree
*******
These commands replaces the sphinx ``toctree`` directive.
A ``toctree`` directive is automatically generated and includes each
page that is a child of the current page.

Example
*******
:ref:`toc_list_example-name`

{xrst_end toc_cmd}
"""
# ---------------------------------------------------------------------------
import os
import xrst
import re
# {xrst_begin toc_cmd_dev dev}
# {xrst_spell
#     toc
# }
# {xrst_comment_ch #}
#
# Get file and page names for children of this page
# #################################################
#
# Arguments
# *********
#
# data_in
# =======
# is the data for the page before the toc commands have been processed.
#
# file_name
# =========
# is the name of the file that this data comes from. This is only used
# for error reporting.
#
# page_name
# =========
# is the name of the page that this data is in. This is only used
# for error reporting.
#
# group_name
# ==========
# We are only including information for pages in this group.
#
# Returns
# *******
#
# data_out
# ========
# is a copy of data_in with the toc commands replaced by \{xrst_command}
# where command is TOC_hidden, TOC_list, or TOC_table depending on
# which command was in data_in.
# There is a newline directly before and after the \{xrst_command}.
#
# file_list
# =========
# is the list of files in the toc command
# (and in same order as in the toc command).
#
# child_page_list
# ===============
# Is the a list of page names corresponding to the children of the
# this page that are in the files specified by file_list.
# If a file in file_list has a begin_parent command, there is only
# one page in child_page_list for that file. Otherwise all of the
# pages in the file are in child_page_list.
#
# {xrst_code py}
def toc_commands(data_in, file_name, page_name, group_name) :
   assert type(data_in) == str
   assert type(file_name) == str
   assert type(page_name) == str
   assert type(group_name) == str
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end toc_cmd_dev}
   #
   # data_out
   data_out = data_in
   #
   # file_list, file_line, child_page_list
   file_list       = list()
   file_line       = list()
   child_page_list = list()
   #
   # m_toc
   m_toc        = xrst.pattern['toc'].search(data_out)
   if m_toc is None :
      xrst.check_syntax_error(
         command_name    = 'toc',
         data            = data_out,
         file_name       = file_name,
         page_name       = page_name,
      )
      return data_out, file_list, child_page_list
   #
   # m_tmp
   m_tmp = xrst.pattern['toc'].search(data_out[m_toc.end() :] )
   if m_tmp is not None :
      msg = 'More than one children or toc_list command in a page.'
      xrst.system_exit(msg,
         file_name=file_name,
         page_name=page_name,
         m_obj=m_tmp,
         data=data_out[m_toc.end():]
      )
   #
   # command
   command = m_toc.group(1)
   assert command in [ 'hidden', 'list', 'table']
   #
   # preceeding_character
   preceeding_character = data_out[ m_toc.start() ]
   assert preceeding_character != '\\'
   #
   # data_out
   replace = preceeding_character + '{xrst_TOC_' + command + '}\n'
   data_out = xrst.pattern['toc'].sub(replace, data_out)
   #
   # file_list, file_line
   child_list =  m_toc.group(2).split('\n')
   child_list = child_list[: -1]
   for child_line in child_list :
      if child_line != '' :
         m_child = xrst.pattern['line'].search(child_line)
         assert m_child != None
         line_number = m_child.group(1)
         child_file  = xrst.pattern['line'].sub('', child_line).strip()
         if child_file != '' :
            file_list.append(child_file)
            file_line.append(line_number)
   #
   # child_page_list
   assert len(child_page_list) == 0
   for i in range( len(file_list) ) :
      #
      # child_file, child_line
      child_file = file_list[i]
      child_line = file_line[i]
      if not os.path.isfile(child_file) :
         msg  = 'The file ' + child_file + ' does not exist\n'
         msg += 'It was used by a toc_' + command + ' command'
         xrst.system_exit(msg,
            file_name=file_name, page_name=page_name, line=child_line
         )
      #
      # child_data
      # errors in the begin and end commands will be detected later
      # when this file is processed.
      file_obj    = open(child_file, 'r')
      child_data  = file_obj.read()
      file_obj.close()
      file_index  = 0
      #
      # m_begin
      m_begin         = xrst.pattern['begin'].search(child_data)
      if m_begin is None :
         msg  = 'The file ' + child_file + '\n'
         msg += 'used in a toc_' + command + ' command does not contain any '
         msg += 'begin commands.'
         xrst.system_exit(msg,
            file_name=file_name, page_name=page_name, line=child_line
         )
      this_group_name = m_begin.group(4).strip(' \t')
      if this_group_name == '' :
         this_group_name = 'default'
      while this_group_name != group_name :
         m_begin = xrst.pattern['begin'].search(child_data, m_begin.end() )
         if m_begin == None :
            msg  = 'The file ' + child_file + '\n'
            msg += 'used in a toc_' + command
            msg += ' command does not contain any '
            msg += f'begin commands with group name {group_name}.'
            xrst.system_exit(msg,
               file_name=file_name, page_name=page_name, line=child_line
            )
         this_group_name = m_begin.group(4).strip(' \t')
         if this_group_name == '' :
            this_group_name = 'default'
      #
      # list_children
      found_parent  = m_begin.group(2) == 'begin_parent'
      child_name    = m_begin.group(3)
      list_children = [ child_name ]
      #
      # m_begin
      m_begin = xrst.pattern['begin'].search(child_data, m_begin.end() )
      #
      while not found_parent and m_begin != None :
         this_group_name = m_begin.group(4).strip(' \t')
         if this_group_name == '' :
            this_group_name = 'default'
         if this_group_name == group_name :
            is_parent       = m_begin.group(2) == 'begin_parent'
            if is_parent :
               msg  = 'Found a begin_parent command that is'
               msg += ' not the first begin command in this file'
               msg += f' for group name {group_name}'
               xrst.system_exit(msg,
                  file_name=child_file,
                  page_name=page_name,
                  m_obj=m_begin,
                  data=child_data
               )
            child_name = m_begin.group(3)
            #
            # list_children
            list_children.append( child_name )
         #
         # m_begin
         m_begin   = xrst.pattern['begin'].search(child_data, m_begin.end() )
      #
      # child_page_list
      child_page_list += list_children
   #
   xrst.check_syntax_error(
      command_name    = 'toc',
      data            = data_out,
      file_name       = file_name,
      page_name       = page_name,
   )
   # We know that the lists are no-empty for this return,
   # but use asserts that work for both retures becase in documentation.
   # BEGIN_return
   assert type(data_out) == str
   assert type(file_list) == list
   if 0 < len(file_list) :
      assert type(file_list[0]) == str
   assert type(child_page_list) == list
   if 0 < len(child_page_list) :
      assert type(child_page_list[0]) == str
   #
   return data_out, file_list, child_page_list
   # END_return
